package jumbo.sgml;
// Jumbo-SAX interface to SAX-based parsers
// Hacked directly from David Megginson's SAXDemo (1998-01-11)
// where possible the comments from David are kept

// $Id: SAXDemo.java,v 1.2 1998/01/11 19:01:11 david Exp $

import org.xml.sax.Parser;
import org.xml.sax.AttributeMap;
import org.xml.sax.EntityHandler;
import org.xml.sax.DocumentHandler;
import org.xml.sax.ErrorHandler;

import java.net.URL;
import java.util.Enumeration;
import java.util.Vector;

import jumbo.simplegraph.SimpleApplet;

/*----
  * Demonstration class for SAX.
  * <p>Locate a SAX Parser class dynamically, based on a command-line
  * argument, use it to parse an XML document, then display
  * the parse events on STDOUT.</p>
  * <p>This demonstration knows only about the SAX interfaces, and
  * has no built-in knowledge of any specific parsers.  It provides
  * a good example of dynamic runtime parser selection.</p>
  * @author Copyright (c) 1998 by Microstar Software Ltd.
  * @author written by David Megginson &lt;dmeggins@microstar.com&gt;
  * @see org.xml.sax.Parser
  * @see org.xml.sax.EntityHandler
  * @see org.xml.sax.DocumentHandler
  * @see org.xml.sax.ErrorHandler
----  */
/** Jumbo-SAX interface class; hacked from SAXDemo
@author Copyright P.Murray-Rust 1998
*/
public class JumboSAX implements EntityHandler, DocumentHandler {

//  final static public String rcsRevision = "$Revision: 1.2 $";

// JUMBO stuff

  DTDTree dtdTree = null;
  SGMLTree tree = null;
  Vector jeVector = new Vector();  // accumulates errors
  static final int MAXERROR = 10;  // beyond this, not reported

  Vector streamVector = new Vector();  // the output
  SGMLAttlist attlist = new SGMLAttlist();
  org.xml.sax.Parser p = null;
  String urlString = null;
  int errorCount;
  ParserDriver parserDriver = null;

/** create a JumboSAX ready for parsing
@param urlString the nascent Tree
@param urlString the URL to parse
@param parserDriver associates parserName (AElfred) with driverName (com.microstar.sax.AelfredDriver)
*/
  public JumboSAX(SGMLTree tree, String urlString, ParserDriver parserDriver)
    throws JumboException {
    if (parserDriver == null) {
        throw new JumboException("Null Parser Driver");
    }
    this.parserDriver = parserDriver;
    String driverName = parserDriver.getDriverName();
    dtdTree = new DTDTree();
    this.tree = tree;
    this.urlString = urlString;
				// Create a new parser.
    try {
      p = createParser(driverName);
    } catch (Exception e) {
      throw new JumboException(e);
    }
				// Register the handler for both
				// entity and basic document events.
    p.setEntityHandler(this);
    p.setDocumentHandler(this);
  }

/** do the parsing and return a streamVector 
@return Vector the result of the parse
*/
  public Vector getSAXStreamVector() throws JumboException {
    try {
      p.parse(null, urlString);
    } catch (org.xml.sax.XmlException e) {
      JumboException je = new JumboException(e, "SAX Error: ");
      je.urlString = e.getSystemID();
      je.line = e.getLine();
      je.column = e.getColumn();
      je.type = JumboException.WHILE_PARSE;
      throw je;
    } catch (Exception e) {
      throw new JumboException(e, "Unknown parse error");
    }
    return streamVector;
  }

  //////////////////////////////////////////////////////////////////////
  // Implement the org.xml.sax.EntityHandler interface.
  //////////////////////////////////////////////////////////////////////


  /**
    * Print a message when resolving an entity.
    * @see org.xml.sax.EntityHandler#resolveEntity
    */
  public String resolveEntity (String ename, String publicID, String systemID)
  {
    System.out.println("Resolving external entity: " + systemID);
    return systemID;
  }


  /**
    * Print a message when the entity changes.
    * @see org.xml.sax.EntityHandler#changeEntity
    */
  public void changeEntity (String systemID)
  {
//    System.out.println("System ID: " + systemID);
  }

  //////////////////////////////////////////////////////////////////////
  // Implement the org.xml.sax.DocumentHandler interface.
  //////////////////////////////////////////////////////////////////////


  /**
    * Print a message at the start of the document.
    * @see org.xml.sax.DocumentHandler#startDocument
    */
  public void startDocument ()
  {
    System.out.println("Start document");
  }


  /**
    * Print a message at the end of the document.
    * @see org.xml.sax.DocumentHandler#endDocument
    */
  public void endDocument ()
  {
    System.out.println("\n***SAX: "+parserDriver+"; End document ***");
  }


  /**
    * Print a message when a document type declaration is found.
    * @see org.xml.sax.DocumentHandler#doctype
    */
  public void doctype (String name, String publicID, String systemID)
  {
    System.out.println("Doctype: " + name + ' ' + publicID + ' ' + systemID);
  }


  /**
    * Print a message at the start of every element.
    * @see org.xml.sax.DocumentHandler#startElement
    */
  public void startElement (String name, AttributeMap atts)
  {
//    System.out.println("Start element: " + name);
    streamVector.addElement("_STARTTAG");
    streamVector.addElement(name);

    Enumeration attNames = atts.getAttributeNames();
    if (attNames == null) {
      System.out.println("[Attributes not available]");
    } else {
      SGMLAttlist attlist = new SGMLAttlist();
      while (attNames.hasMoreElements()) {
        String aname = (String)attNames.nextElement();
        String special = "";
      
        if (atts.isId(aname)) {
	  special = " (ID)";
        } else if (atts.isIdref(aname)) {
	  special = " (IDREF)";
        } else if (atts.isEntity(aname)) {
	  special = " (NDATA entity)";
        } else if (atts.isNotation(aname)) {
	  special = " (notation)";
        }

//      System.out.println("Attribute: " +
//			 aname +
//			 "=\"" +
//			 atts.getValue(aname) +
//			 '"' +
//			 special);
        attlist.addSGMLAttribute(aname, atts.getValue(aname));
      }
      streamVector.addElement(attlist);

    }
  }


  /**
    * Print a message at the end of every element.
    * @see org.xml.sax.DocumentHandler#endElement
    */
  public void endElement (String name)
  {
//    System.out.println("End element: " + name);
    streamVector.addElement("_ENDTAG");
    streamVector.addElement(name);
  }


  /**
    * Print a message when characters are found.
    * @see org.xml.sax.DocumentHandler#characters
    */
  public void characters (char ch[], int start, int length)
  {
//    System.out.println("Characters: \"" +
//		       escapeCharacters(ch, start, length) +
//		       '"');
    streamVector.addElement("#PCDATA");
    streamVector.addElement(new String(ch, 0, length));

  }


  /**
    * Print a message when ignorable whitespace is found.
    * @see org.xml.sax.DocumentHandler#ignorable
    */
  public void ignorable (char ch[], int start, int length)
  {
    System.out.println("Ignorable whitespace: \"" +
		       escapeCharacters(ch, start, length) +
		       '"');
  }


  /**
    * Print a message when a processing instruction is found.
    * @see org.xml.sax.DocumentHandler#processingInstruction
    */
  public void processingInstruction (String name, String remainder)
  {
//    System.out.println("Processing instruction: " +
//		       name + ' ' + remainder);
    dtdTree.addPI(name, remainder);
  }

  //////////////////////////////////////////////////////////////////////
  // Utility routines.
  //////////////////////////////////////////////////////////////////////


  /**
    * Given a class name, create a parser object.
    * <p>This is a fairly involved function, because it requires
    * dynamically loading and instantiating a class based on
    * an arbitrary string containing the class name.</p>
    * <p>The class must implement org.xml.sax.Parser.
    * @param className The name of the class
    * @see org.xml.sax.Parser
    */
  private static org.xml.sax.Parser createParser (String className)
    throws ClassCastException, IllegalAccessException, ClassNotFoundException, InstantiationException
  {
    org.xml.sax.Parser parser = null;

    try {
				// Get the named class.
      Class c = Class.forName(className);
				// Instantiate the parser.
      parser = (org.xml.sax.Parser)(c.newInstance());
    } catch (ClassNotFoundException e) {
      System.err.println("SAX parser class " + className +
			 "cannot be loaded.");
        throw(e);
//      System.exit(1);
    } catch (IllegalAccessException e) {
        throw new IllegalAccessException
          ("SAX parser class " + className + "does not have a zero-argument constructor."+e);
//      System.exit(1);
//    } catch (InstantiationException e) {
//      System.err.println("SAX parser class " + className +
//			 " cannot be instantiated.");
//      System.exit(1);
    }

				// Check the the parser object
				// actually implements the Parser interface.
    if (!(parser instanceof org.xml.sax.Parser)) {
      throw new ClassCastException("Class " + className +
			 " does not implement org.xml.sax.Parser.");
//      System.exit(1);
    }

				// Return the parser.
    return parser;
  }


  /**
    * If a URL is relative, make it absolute against the current directory.
    */
  private static String makeAbsoluteURL (String url)
    throws java.net.MalformedURLException
  {
    URL baseURL;
    String currentDirectory;

    currentDirectory = System.getProperty("user.dir") +
      System.getProperty("file.separator");
    baseURL = new URL("file", "localhost", currentDirectory);
    return new URL(baseURL, url).toString();
  }


  /**
    * Escape special characters for display.
    */
  private static String escapeCharacters(char ch[], int start, int length)
  {
    StringBuffer out = new StringBuffer();

    for (int i = start; i < start+length; i++) {
      if (ch[i] >= 0x20 && ch[i] < 0x7f) {
	out.append(ch[i]);
      } else {
	out.append("&#" + (int)ch[i] + ';');
      }
    }

    return out.toString();
  }

  //////////////////////////////////////////////////////////////////////
  // Entry point for application.
  //////////////////////////////////////////////////////////////////////

  /**
    * Parse an XML document.
    * <p>This program takes two arguments: the first is the 
    * fully-qualified name of the SAX parser (or driver) class, and
    * the second is the URI of the document to parse.
    */
// keep this only for testing - PMR

  public static void main (String args[])
  {
				// Check the command-line arguments.
    if (args.length != 2) {
      System.err.println("Usage: java JumboSAX <parserName> <document>");
      System.exit(1);
    }

				// Create a new parser.
//    Parser p = createParser(args[0]);

				// Create a new handler.
//    SAXDemo demo = new SAXDemo();

				// Register the handler for both
				// entity and basic document events.
//    p.setEntityHandler(demo);
//    p.setDocumentHandler(demo);

    try {
      SGMLTree tree = new SGMLTree();
      JumboConfig jumboConfig = new JumboConfig("jumbo/config.xml");
      ParserDriver parserDriver = new ParserDriver(args[0], jumboConfig);
      String urlString = makeAbsoluteURL(args[1]);
      JumboSAX jumboSAX = new JumboSAX(tree, urlString, parserDriver);
      Vector v = jumboSAX.getSAXStreamVector();
      for (int i = 0; i < v.size(); i++) {
        System.out.println(v.elementAt(i));
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
  }
}
